// Hello World with OpenGL compute shader!
// I have wanted to do this for a long time and now it is done!
// This is slightly less elegant than its siblings since compute shaders
// can not deal with bytes. I chose to repackage to int on the CPU.
// This is an example, use freely.
// By Ingemar Ragnemalm 2017

// Like my similar demos, if it outputs "Hello World!" then it works correctly.
// If it fails, it will most likely output "Hello Hello"

// Compile like this:

// If you use GLUT (see #include below):
// g++ -o hello hello-cs.c -Wall -Iinclude -lGL -lGLU -lX11 -lglut -DGL_GLEXT_PROTOTYPES

// If you use MicroGlut in my typical "common" folder:
// gcc -o hello hello-cs.c ../common/Linux/MicroGlut.c -I ../common/Linux/ -lm -lGL -lX11

// If you use MicroGlut in the same folder:
//gcc -o hello hello-cs.c Linux/MicroGlut.c -I Linux/ -lm -lGL -lX11


#include <stdio.h>
#include <stdlib.h>
#include <sys/times.h>
//#include <GL/glut.h>
#include <GL/gl.h>
#include "MicroGlut.h"

char* readFile(char *file)
{
	FILE *fptr;
	long length;
	char *buf;

	fptr = fopen(file, "rb"); /* Open file for reading */
	if (!fptr) /* Return NULL on failure */
		return NULL;
	fseek(fptr, 0, SEEK_END); /* Seek to the end of the file */
	length = ftell(fptr); /* Find out how many bytes into the file we are */
	buf = (char*)malloc(length+1); /* Allocate a buffer for the entire length of the file and a null terminator */
	fseek(fptr, 0, SEEK_SET); /* Go back to the beginning of the file */
	fread(buf, length, 1, fptr); /* Read the contents of the file in to the buffer */
	fclose(fptr); /* Close the file */
	buf[length] = 0; /* Null terminator */
	
	return buf; /* Return the buffer */
}

// Infolog: Show result of shader compilation
void printShaderInfoLog(GLuint obj, const char *fn)
{
	GLint infologLength = 0;
	GLint charsWritten  = 0;
	char *infoLog;

	glGetShaderiv(obj, GL_INFO_LOG_LENGTH, &infologLength);

	if (infologLength > 2)
	{
		fprintf(stderr, "[From %s:]\n", fn);
		infoLog = (char *)malloc(infologLength);
		glGetShaderInfoLog(obj, infologLength, &charsWritten, infoLog);
		fprintf(stderr, "%s\n",infoLog);
		free(infoLog);
	}
}

void printProgramInfoLog(GLuint obj, const char *vfn)
{
	GLint infologLength = 0;
	GLint charsWritten  = 0;
	char *infoLog;

	glGetProgramiv(obj, GL_INFO_LOG_LENGTH,&infologLength);

	if (infologLength > 2)
	{
		fprintf(stderr, "[From %s:]\n", vfn);
		infoLog = (char *)malloc(infologLength);
		glGetProgramInfoLog(obj, infologLength, &charsWritten, infoLog);
		fprintf(stderr, "%s\n",infoLog);
		free(infoLog);
	}
}
//#include "opengl.h"
//#include <stdio.h>
//#include <stdlib.h>

// Compile a shader, return reference to it
GLuint compileShaders(const char *vs, const char *vfn)
{
	GLuint v,p;
	
	v = glCreateShader(GL_COMPUTE_SHADER); // GL_COMPUTE_SHADER_EXT? GL_COMPUTE_SHADER_ARB?
	glShaderSource(v, 1, &vs, NULL);
	glCompileShader(v);
	
	p = glCreateProgram();//#include "opengl.h"
//#include <stdio.h>
//#include <stdlib.h>

	glAttachShader(p,v);
	glLinkProgram(p);
	glUseProgram(p);
	
	printShaderInfoLog(v, vfn);
	printProgramInfoLog(p, vfn);
	
	return p;
}

GLuint loadShader(char * filename)
{
    char *vs;
       
	vs  = readFile(filename);

	return compileShaders(vs,filename);
}


int main(int argc, char **argv)
{
// Let GLUT create a GL context
	glutInit (&argc, argv);
    glutInitContextVersion(4, 4); // Failed with 4.5 on my PC. The compute shader works even on old-style GL!
	glutCreateWindow("Hello");
    
//	printf ("Vendor: %s\n", glGetString (GL_VENDOR));
//	printf ("Renderer: %s\n", glGetString (GL_RENDERER));
//	printf ("Version: %s\n", glGetString (GL_VERSION));
//	printf ("GLSL: %s\n", glGetString(GL_SHADING_LANGUAGE_VERSION));

// Load and compile the compute shader    
    GLuint p =loadShader("hello.cs");

	GLuint ssbo, ssbo2; //Shader Storage Buffer Object

	// Some data
#define N 16
	char a[N] = "Hello \0\0\0\0\0\0";
	int ac[N];
	int b[N] = {15, 10, 6, 0, -11, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
	int *ptr;
	int i;

	printf("%s", a);

// PROBLEM: No bytes in shaders!
// I chose to package to int on the CPU.
// Convert string to int:
	for (i = 0; i < N; i++) ac[i]=a[i];

// Create buffer, upload data
	glGenBuffers(1, &ssbo);
	glBindBuffer(GL_SHADER_STORAGE_BUFFER, ssbo);
	glBufferData(GL_SHADER_STORAGE_BUFFER, 16 * sizeof(int), &ac, GL_STATIC_DRAW);
	
// Tell it where the input goes!
// The "5" matches a "layuot" number in the shader.
// (Can we ask the shader about the number? I must try that.)
	glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 5, ssbo);

// Same for the other buffer, offsets, ID 6
	glGenBuffers(1, &ssbo2);
	glBindBuffer(GL_SHADER_STORAGE_BUFFER, ssbo2);
	glBufferData(GL_SHADER_STORAGE_BUFFER, 16 * sizeof(int), &b, GL_STATIC_DRAW);
	glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 6, ssbo2);

// Get rolling!
    glDispatchCompute(1, 1, 1); //Work groups launch

// Get data back!
	glBindBuffer(GL_SHADER_STORAGE_BUFFER, ssbo);
	ptr = (int *)glMapBuffer(GL_SHADER_STORAGE_BUFFER, GL_READ_ONLY);
// Convert int to string:
	for (i=0; i < 16; i++)
	{
		a[i] = ptr[i];
	}
	printf("%s\n", a);
		
}

